package engine;
import java.util.*;
import data.*;

// PortGA - java version
// This class should implement the basic calculations and generation methods of a 
// portfolio, so that I can use both GA and GP genomes to generate 
// the same portfolio.

// The portfolio is composed of three parts: A vector of names, which contains
// The name of the composing assets, a vector of weights, and a vector of 
// stock. The input in the portfolio might be either the weights, or the 
// number of each stock.

/* Todo:
 * Portfolio Operation:
 * -- Portfolio Adding
 * -- Portfolio subtracting
 * -- Portfolio weight updating from changed stock value
 * -- rounding off lots based on stock price
 */

public class Portfolio {
	
	private int size;
	private String[] name; 
	private Double[] weight; // weights for each security -- normalized
	private Double[] lots; // ammount of each asset held (usually has to be rounded)
	
	public Portfolio()
	{
		Market mkt = Market.getInstance();
		
		size = mkt.assets.size();
		name = new String[size];
		weight = new Double[size];
		lots = new Double[size];
		
		for (int i = 0; i < size; i++)
			name[i] = mkt.assets.get(i).name;	
	}	
	
	
	/* Defensively copy this portfolio */
	public Portfolio copy()
	{
		Portfolio cp = new Portfolio();
		cp.setLots(this.lots);
		cp.setWeight(this.weight);
		
		return cp;
	}
	
	public int getSize()
	{
		return size;
	}
	
	// Sets the weights values for a portfolio
	// Assumes the wgt array size will be "right"
	// TODO: deal smartly with "wrong" wgt sizes
	public void setWeight(double[] wgt)
	{
		for (int i = 0; i < weight.length; i++)
		{
			weight[i] = wgt[i];
		}
	}
	public void setWeight(Double[] wgt)
	{
		for (int i = 0; i < weight.length; i++)
		{
			weight[i] = wgt[i].doubleValue();
		}
	}
	
	public Double[] getWeight()
	{
		Double[] ret = weight.clone();
		return ret;
	}
	public double getWeight(int i)
	{
		double ret = weight[i];
		return ret;
	}
	
	public void setLots(Double[] l)
	{
		for (int i = 0; i < lots.length; i++)
		{
			lots[i] = l[i].doubleValue();
		}
	}
	public Double[] getLots()
	{
		Double[] ret = lots.clone();
		return ret;
	}
	public double getLots(int i)
	{
		double ret = lots[i];
		return ret;
	}
	
	
	
	// Set sum of weights = 1
	// does not change lots -- need to make internally consistent 
	// after using this.
	public void normalizeWeight()
	{
		double sum = 0;
		for (int i = 0; i < weight.length; i++)
		{
			
			sum += weight[i];
		}
		for (int i = 0; i < weight.length; i++)
		{
			weight[i] = weight[i]/sum;
		}
	}
	
	// Sets the values of the Lots based on the current weights. 
	// Requires a total capital value and a date to be set. It will 
	// take into account the weights of zero-price assets, but the 
	// values of such assets will be zeroed. If an asset is zeroed, 
	// returns false.
	public boolean weightToLots(double c, Date d)
	{
		Market mkt = Market.getInstance();
		boolean zero = true;
		
		for (int i = 0; i < weight.length; i++)
		{
			Double price = mkt.getAssetFromName(name[i]).getPriceByDate(d);
			
			if (price == 0)
			{
				lots[i] = 0.0;
				zero = false;
			}
			else
				lots[i] = (weight[i]*c) / price;
		
		}
			
		return zero;
	}
	
	/*
	 * Calculates a portfolio's weight based on its lots and 
	 * the prices of the assets for a given date.
	 */
	public void lotsToweight(Date d)
	{
		double val = totalValue(d);
		Market mkt = Market.getInstance();
		
		for (int i = 0; i < lots.length; i++)
		{
			weight[i] = (lots[i] * mkt.assets.get(i).getPriceByDate(d)) / val;
		}
		
	}

	/* Returns the total value of the portfolio at a given date. It requires
	 * The number of lots to be set.
	 */
	public Double totalValue(Date d)
	{
		double val = 0.0;
		Market mkt = Market.getInstance();
		
		for (int i = 0; i < lots.length; i++)
		{
			val += lots[i] * mkt.assets.get(i).getPriceByDate(d);
		}
		
		return val;
	}

	/**
	 * 
	 * Generates a String which lists, in decrescent order
	 * the component assets of this Portfolio. 
	 * 
	 * The parameter tresh is the weight treshold for an
	 * asset in the portfolio to be included in the string.
	 * To include all assets (even those with weight 0), 
	 * set tresh to a negative value.
	 * 
	 * @param tresh
	 * @return
	 */
	public String dump(Double tresh)
	{
		String result = new String();
		Vector<PortWeight> pwlist = new Vector<PortWeight>();
			
		for (int i = 0; i < size; i++)
		{
			pwlist.add(new PortWeight(name[i],weight[i]));
		}
		
		Collections.sort(pwlist);

		int i = 0;
		while(i < pwlist.size() && pwlist.get(i).weight > tresh)
		{
			result = result + pwlist.get(i).name + ":\t" + pwlist.get(i).weight + "\n";
			i++;
		}
		
		return result;
	}
}
