
package gp;
/* 
 * Population
 * 
 * This class implements a GA optimization of a group of investment weights over some return data. 
 * You must provide the data and some GA parameters, and then you can run the GA one generation at once, 
 * for a set number of generations, or until the fitness converges.
 * 
 * 
 */

import engine.*;
import java.util.*;

public class GPPopulation {

	/* Evolutionary parameters: */
	int size = 10; // size of the population
	int ngens = 10; // number of generations
	int tournamentK = 5; // number of individuals chosen to make each tournament
	int elite = 2; // number of individuals kept from the previous generation to the next
	int immigrant = 2; // number of new random individuals
	double mutrate = 0.05; // chance that a mutation will occur
	double xoverrate = 0.8; // chance that the crossover will occur
	int MOGAfactor = 1; // what kind of MOGA to use : 1 - no MOGA
	
	/* GP Parameters */
	int maxdepth = 8; // Maximum depth for the tree.
	int fulltree = 1; // true/false. Are we dealing only on full trees?
	int subtree_fit = -1; // -1 is no subtree fit
	// 0 is Sharpe, 1 is Areturn, 2 is Creturn, 3 is risk;
	
	Double AssTresh = 0.00001;
	
	/* problem parameters */
	// A GP run is usually done for one scenario.
	Date date = null;
	int moving_average = 12;
	
	// Riskless return rate. Should be parametrized, but if it isn't,
	// set the parameter to this value.
	Double riskless_return = 0.02;
	
	/* Containers */
	ArrayList<Node> individual;
	Node parent; // the previous portfolio, used for comparison - might be null if there's no previous port.
		
	/* Local variables */
	private boolean evalDone = false; // have we already evaluated the current population?
	
	// Constructor - basic, does not create a generation.
	public GPPopulation(Node comp)
	{
		Parameter param = Parameter.getInstance();
		
		String paramval;
		individual = new ArrayList<Node>();
		parent = comp;

		/* setting parameters */
		
		paramval = param.getParam("population size");
		if (paramval != null)
			size = Integer.valueOf(paramval);

		paramval = param.getParam("generation number");
		if (paramval != null)
			ngens = Integer.valueOf(paramval);
		
		paramval = param.getParam("tournament K");
		if (paramval != null)
			tournamentK = Integer.valueOf(paramval);
		
		paramval = param.getParam("elite size");
		if (paramval != null)
			elite = Integer.valueOf(paramval);

		paramval = param.getParam("immigrant size");
		if (paramval != null)
			immigrant = Integer.valueOf(paramval);

		paramval = param.getParam("mutation rate");
		if (paramval != null)
			mutrate = Double.valueOf(paramval);

		paramval = param.getParam("MOGA factor");
		if (paramval != null)
			MOGAfactor = Integer.valueOf(paramval);

		paramval = param.getParam("Max Depth");
		if (paramval != null)
			maxdepth = Integer.valueOf(paramval);
		
		paramval = param.getParam("Full Tree");
		if (paramval != null)
			fulltree = Integer.valueOf(paramval);
				
		// This one is different. It sets the Riskless Return if it 
		// is not set.
		paramval = param.getParam("Riskless Return");
		if (paramval != null)
			riskless_return = Double.valueOf(paramval);

		paramval = param.getParam("Subtree Fitness");
		if (paramval != null)
			subtree_fit = Integer.valueOf(paramval);
		
		paramval = param.getParam("Moving Average");
		if (paramval != null)
			moving_average = Integer.valueOf(paramval);
		
		paramval = param.getParam("Asset Treshold");
		if (paramval != null)
			AssTresh = Double.valueOf(paramval);
	
	}
	
	/**
	 * Sets the date for the scenario that is going to be solved. 
	 * 
	 * The moving average is calculated between these dates - 
	 * so they must be ideally at least a month apart, maybe more.
	 * 
	 */
	public void setProbParam(Date startd)
	{
		date = startd;
		evalDone = false;
	}
	
	/**
	 * initPop: this method initializes the population. 
	 * If called with a parameter, copy the N best individuals 
	 * from the previous state of the population. Should be 
	 * able to deal with empty populations in any cases.
	 * 
	 */
	public void initPop()
	{
		initPop(0);
	}
	public void initPop(int N)
	{
		
		ArrayList<Node> newpop = new ArrayList<Node>();

		if (N < 0)
		{
			System.out.println("ERROR Population.java: Invalid Parameter - reset with negative N.");
			System.exit(1);
		}
		
		// If the population is not empty, copy the best individual to the parent
		if (!individual.isEmpty())
		{
			eval();
			parent = individual.get(0).copy();
		}
			
		// first copy the old individuals
		if (N > 0)
		{
			eval();
			for (int i = 0; i < N; i++)
				newpop.add(individual.get(i).copy());
		}
		// Now generate the random individuals
		for (int i = N; i < size; i++)
		{
			Node newn = new Node();
			newn.generateFullTree(maxdepth);
			newpop.add(newn);
		}
		
		individual = newpop;
		
		evalDone = false;
		
	}
	


	
	/**
	 * Evolve: runs a number of generations specified in the 
	 * object parameters. 
	 */
	public void evolve()
	{
		// fitness values
		double[] maxfitness = new double[ngens];
		double[] avgfitness = new double[ngens];
		// total asset values
		double[] maxasset = new double[ngens];
		double[] avgasset = new double[ngens];
		// total assets above a given treshold
		double[] maxbigasset = new double [ngens];
		double[] avgbigasset = new double [ngens];
		
		for (int i = 0; i < ngens; i++)
		{
			
			if (i%MOGAfactor == 1)
				eval(1);
			else
				eval(0);
							
			maxfitness[i] = individual.get(0).fitness;
			maxasset[i] = individual.get(0).countAsset(0);
			maxbigasset[i] = individual.get(0).countAsset(AssTresh);
			
			for (int j = 0; j < individual.size(); j++)
			{
				avgfitness[i] += individual.get(j).fitness;
				avgasset[i] += individual.get(j).countAsset(0);
				avgbigasset[i] += individual.get(j).countAsset(AssTresh);
			}
			avgfitness[i] /= individual.size();
			avgasset[i] /= individual.size();
			avgbigasset[i] /= individual.size();

			/* Prints information about the generation results.
			 * TODO: implement a switch to turn on/off the display
			 * of this information.
			 */
			//System.out.print(i + " " + 
			//		maxfitness[i] + " " + avgfitness[i] + " " +
			//		maxasset[i] + " " + avgasset[i] + " " +
			//		maxbigasset[i] + " " + avgbigasset[i] + "\n");
			
			breed();		
		}
		
	}
	
	/**
	 * Eval: evaluates and sorts the individuals in the 
	 * population. Register that it has already done the evaluation 
	 * so that it doesn't do it again if not needed.]
	 */
	public void eval()
	{
		eval(0);
	}
	public void eval(int fitmode)
	{
		if (!evalDone)
		{
			// If dateinit and dateend values are not set, bails.
			if (date == null)
			{
				System.err.print("Error - called Population.eval without setting the date beforehand. Bailing out\n");
				System.exit(1);
			}
			
			Calendar c = Calendar.getInstance();
			c.setTime(date);
			// Fitness of the portfolio is actually tested on the month 
			// PRIOR to the scenario month - since we use only past 
			// data.
			c.add(Calendar.MONTH, -1);
			Date curdate = c.getTime();
			c.add(Calendar.MONTH, -1*moving_average);
			Date pastdate = c.getTime();
			
			for (int i = 0; i < individual.size(); i++)
			{
				
				individual.get(i).calcRecursiveResult(pastdate, curdate);
				
				/* TODO: Select a more appropriate fitness 
				 * according to the experiment. Here I'm 
				 * always using Sharpe Ratio
				 * (this should be enough for GECCO)
				 */
				individual.get(i).setFitness(individual.get(i).Sharpe);
			}
			
			Collections.sort(individual);
		}
		
		evalDone = true;	
	}
	
		
	// Breeds a new population, using current parameters. One must remember to eval the population first.
	public void breed()
	{
		
		RNG die = RNG.getInstance();
		ArrayList<Node> nextGen = new ArrayList<Node>();
		int p1, p2;
		// Elite - copy the best old individuals
		for (int i = 0; i < elite; i++)
		{
			nextGen.add(individual.get(i).copy());
		}
		// Immigration - new random individuals - shouldn't use this much.
		for (int i = 0; i < immigrant; i++)
		{
			Node newn = new Node();
			newn.generateFullTree(maxdepth);
			nextGen.add(newn);
		}
		
		// Using deterministic tournament selection. Maybe I 
		// should use stochastic tournament?
		for (int i = 0; i < size - (elite + immigrant); i+=2)
		{
			p1 = tournament();
			p2 = tournament();
			
			Node n1 = individual.get(p1).copy();
			Node n2 = individual.get(p2).copy();

			/* TODO: add xover counters */
			if (die.nextDouble() < xoverrate)
			{
				if (subtree_fit == -1)
					n2.crossoverFullTree(n1);
				else
					n2.crossoverFTreeGuidedBestWorst(n1, subtree_fit);
			}

			/* TODO: Add mutation counters */
			if (die.nextDouble() < mutrate)
				n1.mutateFullTree();
			if (die.nextDouble() < mutrate)
				n2.mutateFullTree();
			
			nextGen.add(n1);
			nextGen.add(n2);
						
		}
				
		individual = nextGen;
		evalDone = false;
		/* TODO: Don't I have to delete the 
		 * previous generation first?
		 */
		
	}

	/**
	 * returns the best N individuals in a list, with 
	 * their fitness values already filled out.
	 * 
	 * @param N
	 * @return
	 */
	public Node[] bestIndiv(int N)
	{
		Node[] result = new Node[N];

		// sorts the population
		eval();
		
		for (int i = 0; i < N; i++)
		{
			// clones the victors so that we may alter them;
			result[i] = individual.get(i).copy();
		}

		return result;
	}

	/**
	 * Chooses one individual by deterministic tournament
	 * 
	 * @return the index of the selected parent
	 */
	private int tournament()
	{
		RNG dice = RNG.getInstance();
		
		int victor = dice.nextInt(individual.size());
		int t0;
		for (int i = 0; i < tournamentK - 1; i++)
		{
			t0 = dice.nextInt(individual.size());
			victor = (victor < t0?victor:t0);
		}
		return victor;
	}
	
	
}
