/**
 * The node is the base of the tree genome. Can be a real node or a Terminal.
 */

/* FIXME: add a destroy method? */

package gp;

import data.*;
import engine.*;

import java.util.*;

public class Node implements Comparable<Node>{
	
	
	//	 characteristics of the node	
	Boolean isTerminal;	
	Long uniqueID;
	Integer depth; // current level. A tree with d = 0 is uninitialized.
	Integer treedepth; // Depth of the deepest level of the tree.
	
	
	// Contents of the node
	Node leftChild;
	Node rightChild;
	Double weight; // weight of the children, if node
	Integer assetIndex; // index of the asset, if terminal
	
	
	// Fitness value of this subtree
	
	public Double fitness = 0.0;
	public Double Areturn = 0.0; // average return
	public Double Creturn = 0.0; // cumulative return
	public Double Risk    = 0.0; // risk measure
	public Double Sharpe  = 0.0; // sharpe ratio
	
	
	/**
	 * Empty constructor - generates an empty node.
	 *
	 */
	public Node()
	{		
		IDGenerator i = IDGenerator.getInstance();
		uniqueID = i.newID();
		depth = 0;
	}
	
	/**
	 * Copy tree function - recursively copies a tree and send 
	 * it back
	 */
	public Node copy()
	{
		if (isTerminal)
		{
			Node newn = new Node();
			newn.depth = depth;
			newn.treedepth = treedepth;
			newn.isTerminal = true;
			newn.assetIndex = assetIndex;
			return newn;
		}
		else
		{
			Node newn = new Node();
			newn.depth = depth;
			newn.treedepth = treedepth;
			newn.isTerminal = false;
			newn.weight = weight;
			newn.leftChild = leftChild.copy();
			newn.rightChild = rightChild.copy();
			return newn;
		}
	}

	/**
	 * Generate a full random tree starting from this node. Recursive Function.
	 * A tree with only the root has depth = 0. The correct depth of the node is 
	 * supposed to be already set when entering this function.
	 */
	public void generateFullTree(int maxdepth)
	{
		RNG dice = RNG.getInstance();
		treedepth = maxdepth;
		if (depth == 0)
			depth = 1;
			
		if (maxdepth == depth)
		{
			// terminal case
			Market mkt = Market.getInstance();
			isTerminal = true;
			assetIndex = dice.nextInt(mkt.assets.size());			
		}
		else
		{
			// node case
			weight = dice.nextDouble();
			isTerminal = false;
			leftChild = new Node();
			leftChild.depth = depth + 1;
			leftChild.generateFullTree(maxdepth);
			rightChild = new Node();
			rightChild.depth = depth + 1;
			rightChild.generateFullTree(maxdepth);			
		}
	}

	/**
	 * Crossover this tree with another similar full tree.
	 * This is a simple crossover where nodes in the same 
	 * level are chosen for changing, randomly. The 
	 * tree depths are not changed. The parameter tree
	 * and the current tree are modified.
	 */
	public void crossoverFullTree(Node tree)
	{
		boolean myleft, hisleft; /* if we're switching the left nodes */
		Node mypnode, hispnode; /* the parent nodes */
		Node mycnode, hiscnode; /* the nodes being xovered */
		
		RNG die = RNG.getInstance();
		int cutdepth = die.nextInt(treedepth -1); 
		cutdepth ++; // don't crossover at root.

		// ugly code to avoid special first case
		mycnode = this;
		hiscnode = tree;
		int i = 0;
		
		// choose cutpoints
		do {
			mypnode = mycnode;
			hispnode = hiscnode;
			myleft = die.nextBoolean();
			hisleft = die.nextBoolean();
			
			mycnode = (myleft?mypnode.leftChild:mypnode.rightChild);
			hiscnode = (hisleft?hispnode.leftChild:hispnode.rightChild);
			
			i++;
		} while (i < cutdepth);
		
		// transplant subtrees.
		
		if (myleft) 
			mypnode.leftChild = hiscnode;
		else
			mypnode.rightChild = hiscnode;
		
		if (hisleft)
			hispnode.leftChild = mycnode;
		else
			hispnode.rightChild = mycnode;
		
	}

	/**
	 * Basic Guided Crossover. Exchange the WORST subtree of 
	 * this tree with the BEST subtree of the parameter tree. 
	 * This function changes the two trees, and assumes that 
	 * each tree has gone through the RecursiveReturn function.
	 * 
	 * @param tree The tree we are crossovering with.
	 * @param type 0- Sharpe 1-Av. Return 2-Cum. Return 3-Risk
	 */
	public void crossoverFTreeGuidedBestWorst(Node tree, int type)
	{
		Boolean myleft = null, hisleft = null; /* if we're switching the left nodes */
		Node mypnode = null, hispnode = null; /* the parent nodes */
		Node mycnode = null, hiscnode = null; /* the nodes being xovered */
	
		RNG die = RNG.getInstance();
		
		int cutdepth = die.nextInt(treedepth -1); 
		cutdepth ++; // don't crossover at root.
		
		// Choosing cutpoints:
		// mypnode, myleft, mycnode
		// hispnode, hisleft, hiscnode
		
		
		// Select all parents.
		int nparents = (int) Math.pow(2, cutdepth-1);
		Node myparents[] = new Node[nparents];
		Node hisparents[] = new Node[nparents];
		
		myparents[0] = this;
		hisparents[0] = tree;
		int i = 0;
		int j = 1;
				
		while (j < nparents*2 - 1)
		{
			myparents[j%nparents] = myparents[i%nparents].leftChild;
			myparents[(j+1)%nparents] = myparents[i%nparents].rightChild;
			hisparents[j%nparents] = hisparents[i%nparents].leftChild;
			hisparents[(j+1)%nparents] = hisparents[i%nparents].rightChild;
			
			i +=1;
			j +=2;
		}

		// Select the best and children
		
		//System.out.print("\n Cut depth" + cutdepth + "\n");
		
		double myMinvalue = Double.POSITIVE_INFINITY;
		double hisMaxvalue = Double.NEGATIVE_INFINITY;

		for (i = 0; i < nparents; i++)
		{
			switch (type)
			{
			case 0: // Sharpe
			// evaluate left
			//System.out.print(myparents[i].leftChild.Sharpe + " " + hisparents[i].leftChild.Sharpe + "\n");
			if (myparents[i].leftChild.Sharpe < myMinvalue)
			{
				
				myMinvalue = myparents[i].leftChild.Sharpe;
				mycnode = myparents[i].leftChild;
				mypnode = myparents[i];
				myleft = true;
			}			
			if (hisparents[i].leftChild.Sharpe > hisMaxvalue)
			{
				hisMaxvalue = hisparents[i].leftChild.Sharpe;
				hiscnode = hisparents[i].leftChild;
				hispnode = hisparents[i];
				hisleft = true;
			}
			
			// evaluate right child
			// System.out.print(myparents[i].rightChild.Sharpe + " " + hisparents[i].rightChild.Sharpe + "\n");
			if (myparents[i].rightChild.Sharpe < myMinvalue)
			{
				myMinvalue = myparents[i].rightChild.Sharpe;
				mycnode = myparents[i].rightChild;
				mypnode = myparents[i];
				myleft = false;
			}			
			if (hisparents[i].rightChild.Sharpe > hisMaxvalue)
			{
				hisMaxvalue = hisparents[i].rightChild.Sharpe;
				hiscnode = hisparents[i].rightChild;
				hispnode = hisparents[i];
				hisleft = false;
			}
			break;
			case 1: //Areturn
			case 2: //Creturn
			case 3: //Risk
				break;	
			
			}
			
		}
		
		
		// transplant subtrees.
		
		if (myleft == null||hisleft == null)
		{
			System.err.print("Escaping bad crossover. Verify Guided Crossover Function");
			return;
		}
		
		if (myleft) 
			mypnode.leftChild = hiscnode;
		else
			mypnode.rightChild = hiscnode;
		
		if (hisleft)
			hispnode.leftChild = mycnode;
		else
			hispnode.rightChild = mycnode;
		

	}
	
	/**
	 * Mutate a full tree with a simple full subtree mutation.
	 * a random point in the tree will be chosen, and a 
	 * full subtree will be re-generated at that point.
	 */
	public void mutateFullTree()
	{
		RNG die = RNG.getInstance();
		int cutdepth = die.nextInt(treedepth); // may mutate at root
		Node p = this;
		
		for (int i = 0; i < cutdepth; i++)
		{
			if (die.nextBoolean())
				p = p.leftChild;
			else 
				p = p.rightChild;
		}
		
		/* TODO: maybe I should delete the tree first? */
		p.generateFullTree(p.treedepth);
	}
	
	/**
	 * Recursively Generates a 
	 * array of double weights for the portfolio
	 */	
	public double[] GeneratePortfolio()
	{
		Market mkt = Market.getInstance();

		double[] result = new double[mkt.assets.size()];
		
		if (isTerminal)
			{
				result[assetIndex] = 1.0;
			}
		else
		{
			double[] lresult = leftChild.GeneratePortfolio();
			double[] rresult = rightChild.GeneratePortfolio();	
			for (int i = 0; i < result.length; i++)
			{
					
				result[i] = (lresult[i] * weight) + (rresult[i] * (1 - weight));
			}
		}
		
		return result;
		
	}
	
	/**
	 * Getter for fitness of this individual
	 * @return
	 */
	public Double getFitness()
	{
		return fitness;
	}
	
	/**
	 * Setter for fitness of this individual
	 */
	public void setFitness(Double f)
	{
		fitness = f;
	}
	
	
	/** 
	 * Removes invalid assets from the tree, pruning them.
	 * runs the tree, looking for end nodes that don't contain price 
	 * values for the given time period, and remove them from the 
	 * tree. Need to change weight and/or tree structure.
	 * 
	 * Limit is a parameter that says what percentage of the prices 
	 * need to be valid for the asset to be valid. If set to 1, 
	 * all prices in the period must be valid. If set lower, a 
	 * proportionally higher tolerance to invalid prices is used.
	 * 
	 * Usually you should always use 1 as the treshold.
	 * 
	 */
	public void removeInvalidAssets(Date startd, Date Endd, double limit)
	{		
	}

	/**
	 * Prunes the tree to the maximum size. The subtrees pruned 
	 * will be reduced to the asset with the biggest total weight of 
	 * the subtree.
	 * 
	 * @param size maximum size of the tree.
	 */
	public void pruneTree(int size)
	{
		
	}
	
	
	/**
	 *  Recursively calculates and sets the value of the Risk 
	 * and the Returns for this tree and all its subtrees - 
	 * Returns a vector with the returns which is used to quickly 
	 * calculate the return for the next tree.
	 *
	 * CAVEAT: This function does not take into account changes
	 * to the portfolio or rebalancing costs. It does the return
	 * calculation by itself, without using the Trader object.
	 */
	public Double[] calcRecursiveResult(Date init, Date end)
	{
		Market m = Market.getInstance();
		int time = Utils.calcMonths(init,end);

		Double[] ret = new Double[time];
		Double cur,prev;
		
		Creturn = 1.0;
		Areturn = 0.0;
		
		/* First, calculate the return */
		if (isTerminal)
		{
			int t0 = m.assets.get(assetIndex).getIndexByDate(init);			
			cur = m.assets.get(assetIndex).getPriceByIndex(t0);
			int i = 0;
			
			do
			{
				prev = cur;
				cur = m.assets.get(assetIndex).getPriceByIndex(t0 + i + 1);
				
				/* Treatment of NaN Cases:
				 * NaNs happen when the asset is not available at 
				 * a given time period. If that is the case for terminals,
				 * the return is 0. The fitness function should try to 
				 * avoid assets with the return = 0;
				 * 
				*/
				if (prev == 0.0 || cur == 0.0)
				{
					ret[i] = 0.0;
				}
				else
				{
					ret[i] = Math.log(cur/prev);				
				}
				
				Creturn *= ret[i] + 1;
				Areturn += ret[i];
				i++;				
			} while (i < time);
		}
		else
		{
			Double[] rval = this.rightChild.calcRecursiveResult(init, end);
			Double[] lval = this.leftChild.calcRecursiveResult(init, end);
			for (int i = 0; i < time; i++)
			{
				
				ret[i] = lval[i]*weight + rval[i]*(1-weight);
				Creturn *= ret[i] + 1;
				Areturn += ret[i];
			}
		}
		
		
		Parameter p = Parameter.getInstance();
		Double fixed = Double.valueOf(p.getParam("riskless return"));
		
		Areturn /= ret.length;
		Risk = Utils.variance(ret);	
		Sharpe = (Areturn - fixed)/Risk;
		
		return ret;
	}
	
	/**
	 * Calculates and sets up the risk and the return of this 
	 * Tree, without calculating the values of its children.
	 * 
	 * This function uses the Trader object, so it SHOULD be 
	 * able to take into accound rebalancing and rebalancing 
	 * costs, if the Trader does that correctly.
	 *
	 */
	public void calcLocalRiskReturn(Date init, Date end)
	{
		
	}
	
	/* implements comparable. Higher fitness is better.
     */
    public int compareTo(Node gt)
    {
    	if (getFitness() > gt.getFitness())
    		return -1;
    	if (getFitness() < gt.getFitness())
    		return 1;
    	return 0;
    
    }
    
    public int countAsset(double tresh)
    {
    	int ret = 0;
    	double[] w = GeneratePortfolio();
    	for (int i = 0; i < w.length; i++)
    		if (w[i] > tresh) 
    			ret++;
    	return ret;    		
    }
	
}
