package memetic;

import data.*;
import engine.*;


public class MemeNode implements Comparable<MemeNode> {

	/* node internal variables */
	Boolean isTerminal;	
	Long uniqueID;
	int mydepth; // depth from root.
	int treedepth; // depth of this subtree.
	MemeNode[] best_subtree; // one per depth from here.
	MemeNode[] worst_subtree; // one per depth from here
	Boolean sane; /* are the above arrays sane? */
	
	/* Node Contents */
	public MemeNode leftChild;
	public MemeNode rightChild;
	Double weight; // weight of the children, if node
	Integer assetIndex; // index of the asset, if terminal
	
	/* Fitness Values */
	
	public Double fitness = 0.0;
	public Double Ereturn = 0.0; // estimated return
	public Double Risk    = 0.0; // risk measure
	public Double Sharpe  = 0.0; // sharpe ratio
	
	/**
	 * Empty constructor - generates an empty node.
	 *
	 */
	public MemeNode()
	{		
		IDGenerator i = IDGenerator.getInstance();
		uniqueID = i.newID();
		mydepth = -1; //this node is uninitiated;
	}

	/**
	 * Copy tree function - recursively copies a tree and send 
	 * it back
	 */
	public MemeNode copy()
	{
		IDGenerator i = IDGenerator.getInstance();
		if (isTerminal)
		{
			MemeNode newn = new MemeNode();
			newn.mydepth = mydepth;
			newn.treedepth = treedepth;
			newn.isTerminal = true;
			newn.assetIndex = assetIndex;
			newn.sane = false;
			newn.uniqueID = i.newID();
			newn.fitness = fitness;
			return newn;
		}
		else
		{
			MemeNode newn = new MemeNode();
			newn.mydepth = mydepth;
			newn.treedepth = treedepth;
			newn.isTerminal = false;
			newn.weight = weight;
			newn.leftChild = leftChild.copy();
			newn.rightChild = rightChild.copy();
			newn.uniqueID = i.newID();
			newn.sane = false;
			newn.fitness = fitness;
			return newn;
		}
	}
	
	/**
	 * Generate a random tree starting from this node. 
	 * Recursive Function. A tree with only the root 
	 * has depth = 0. The correct depth of the node is 
	 * supposed to be already set when entering this 
	 * function (-1 if empty tree).
	 * 
	 * prob is the probability that each node will 
	 * not be a terminal node. If prob = 1, then the 
	 * tree is a full tree.
	 * 
	 */
	public void generateTree(int maxdepth, double prob)
	{
		RNG dice = RNG.getInstance();
		if (mydepth == -1)
			mydepth = 0;
			
		if ((maxdepth == mydepth+1)||
				(dice.nextDouble() > prob))
		{
			// terminal case
			Market mkt = Market.getInstance();
			treedepth = 1;
			isTerminal = true;
			assetIndex = dice.nextInt(mkt.assets.size());			
			sane = false;
		}
		else
		{
			// node case
			weight = dice.nextDouble();
			isTerminal = false;
			sane = false;
			leftChild = new MemeNode();
			leftChild.mydepth = mydepth + 1;
			leftChild.generateTree(maxdepth,prob);
			
			rightChild = new MemeNode();
			rightChild.mydepth = mydepth + 1;
			rightChild.generateTree(maxdepth,prob);	
			
			treedepth = (leftChild.treedepth > rightChild.treedepth ?
					leftChild.treedepth:rightChild.treedepth);
			treedepth+=1;
		}
	}
	
	/**
	 * Crossover this tree with another tree at random
	 * points. Depth of the cut is decided with Normal
	 * probability for both trees, and the actual node 
	 * is chosen by visiting the tree.
	 * 
	 * If each resulting tree is deeper than maximum 
	 * depth, the tree is pruned.
	 */
	public void crossoverSimple(MemeNode tree)
	{
		boolean myleft, hisleft; /* if we're switching the left nodes */
		MemeNode mypnode, hispnode; /* the parent nodes */
		MemeNode mycnode, hiscnode; /* the nodes being xovered */
		int mycut, hiscut, i; /* cut depths */
		
		RNG die = RNG.getInstance();
		
		// do not crossover unitary trees
		if (this.treedepth == 1 || tree.treedepth == 1)
			return;

		// ugly code to avoid special first case
		mycnode = this;
		hiscnode = tree;

		mycut = 1 + die.nextInt(this.treedepth-1);
		hiscut = 1 + die.nextInt(tree.treedepth-1);
		
		// choose cutpoints
		
		i = 0;
		do {
			mypnode = mycnode;
			mypnode.sane = false;
			myleft = die.nextBoolean();			
			mycnode = (myleft?mypnode.leftChild:mypnode.rightChild);			
			i++;
		} while ((i < mycut) && (mycnode.isTerminal==false));

		i = 0;
		do {
			hispnode = hiscnode;
			hispnode.sane = false;
			hisleft = die.nextBoolean();
			hiscnode = (hisleft?hispnode.leftChild:hispnode.rightChild);			
			i++;
		} while ((i < hiscut) && (hiscnode.isTerminal==false));
		
		
		// transplant subtrees.
		
		if (myleft) 
			mypnode.leftChild = hiscnode;
		else
			mypnode.rightChild = hiscnode;
		
		if (hisleft)
			hispnode.leftChild = mycnode;
		else
			hispnode.rightChild = mycnode;	
		
		tree.mydepth = 0;
		this.mydepth = 0;
		
		if (!(this.insanity_Check() && tree.insanity_Check()))
		{
			System.err.println("insane before pruning");
		}
		
		tree.prune();
		this.prune();

		
	}

	/**
	 * Crossover this tree with another tree at 
	 * points determined by the best and worst vector.
	 * Depth is chosen randomly, and the best 
	 * subtree in "tree" is crossed over with the
	 * worst subtree in "this".
	 * 
	 * Prunes the tree later.
	 * 
	 * @param tree
	 */
	public void crossoverGuided(MemeNode tree)
	{
		boolean myleft, hisleft; /* if we're switching the left nodes */
		MemeNode mypnode, hispnode; /* the parent nodes */
		MemeNode mycnode, hiscnode; /* the nodes being xovered */
		int cut; /* cut depths */
		
		RNG die = RNG.getInstance();

		int depth = (this.treedepth > tree.treedepth?tree.treedepth:this.treedepth);

		if (depth == 1) 
			return;// do not crossover trees with size = 1;
		
		cut = 1 + die.nextInt(depth-1);

		
		/*
		 * The cut points are decided by the best/worst list.
		 * for each level, finds out whether to go left or
		 * right.
		 */
		mycnode = worst_subtree[cut];
		hiscnode = tree.best_subtree[cut];
		
		// special first case;
		mypnode = this;
		hispnode = tree;
		myleft = true;
		hisleft = false;


		do 
		{			
			mypnode.sane = false;
			hispnode.sane = false;
			if (mypnode.leftChild.worst_subtree.length > cut-1 &&
				mypnode.leftChild.worst_subtree[cut-1].uniqueID ==
				mycnode.uniqueID)
			{
				if (cut > 1)
					mypnode = mypnode.leftChild;
				else
					myleft = true;
			}
			else
			{
				if (cut > 1)
					mypnode = mypnode.rightChild;
				else
					myleft = false;
			}
			
			if (hispnode.leftChild.best_subtree.length > cut - 1 &&
				hispnode.leftChild.best_subtree[cut-1].uniqueID ==
				hiscnode.uniqueID)
			{
				if (cut > 1)
					hispnode = hispnode.leftChild;
				else
				{					
					hisleft = true;
				}
			}
			else
			{
				if (cut > 1)
					hispnode = hispnode.rightChild;
				else
				{
					hisleft = false;
				}
			}

			cut --;
		} while (cut > 0);		

		
		// transplant subtrees.
		
		if (myleft) 
			mypnode.leftChild = hiscnode;
		else
			mypnode.rightChild = hiscnode;
		
		if (hisleft)
			hispnode.leftChild = mycnode;
		else
			hispnode.rightChild = mycnode;	
		
		tree.mydepth = 0;
		this.mydepth = 0;
		tree.prune();
		this.prune();		
	}

	
	/**
	 * Mutation chooses a node at random in the tree, 
	 * and generates a new tree based on that node.
	 *
	 */
	public void mutation()
	{
		RNG die = RNG.getInstance();
		Parameter p = Parameter.getInstance();
		String param = p.getParam("tree depth");
		if (param == null)
		{
			System.err.println("Error in Mutation: No tree depth parameter defined");
			System.exit(1);
		}
		int maxdepth = Integer.parseInt(param);	
		
		param = p.getParam("tree density");
		double dense = Double.parseDouble(param);
		
		int cutdepth = die.nextInt(treedepth); // may mutate at root
		MemeNode n = this;
		
		for (int i = 0; i < cutdepth; i++)
		{
			sane = false; // because this node's kids changed.
			if (n.isTerminal)
				break;
			if (die.nextBoolean())
				n = n.leftChild;
			else 
				n = n.rightChild;
		}
		
		
		/* TODO: maybe I should delete the tree first? */
		n.generateTree(maxdepth, dense);
		prune();
	}
	
	/**
	 * Walks through the trees, correcting the 
	 * mydepth values for the children. If mydepth
	 * reaches the maxdepth parameter value, prunes 
	 * the tree.
	 * 
	 * TODO: This function is inneficient, as 
	 * it runs through the whole tree. I need to 
	 * fix this object as to only visit the 
	 * nodes that were changed in a tree.
	 * 
	 */
	public void prune()
	{
		Parameter p = Parameter.getInstance();
		int maxdepth = Integer.parseInt(p.getParam("tree depth"));
		
		// If this node is a terminal, do nothing.
		if (isTerminal == true)
		{
			treedepth = 1;
			return;
			
		}
			
		if (isTerminal == false && mydepth+1 < maxdepth)
		{
			// Non terminal, and below depth limit.
			// Just call this function recursively.
			leftChild.mydepth = mydepth + 1;
			leftChild.prune();
			rightChild.mydepth = mydepth + 1;
			rightChild.prune();
			
			// fix sanity and tree depths.
			treedepth = (leftChild.treedepth > rightChild.treedepth ?
					leftChild.treedepth:rightChild.treedepth);
			treedepth+=1;
			if (!leftChild.sane || !rightChild.sane)
				sane = false;
			
		}
		else
		{
			/* PRUNING HAPPENS HERE.
			 * 
			 * Pruning policy checks the highest weighted assed
			 * of this subtree and transforms it into a 
			 * terminal of that asset.
			 */
			double[] weights = this.generatePortfolio();
			int maxindex = 0;
			for (int i = 1; i < weights.length; i++)
			{
				if (weights[i] > weights[maxindex])
					maxindex = i;
			}
	
			treedepth = 0;
			sane = false;
			
			worst_subtree = null;
			best_subtree = null;
			
			isTerminal = true;
			assetIndex = maxindex;
			
			/* TODO: maybe I should finalize these? */
			leftChild = null;
			rightChild = null;
			
			
		}
	}
	
	/**
	 * Runs recursively through the tree, 
	 * optimizing each non-terminal weight value.
	 * 
	 * TODO: implement the sanity bit correctly, 
	 * in order to reduce the time complexity of 
	 * this (and other) functions;
	 */
	public void memetize(Trainer T)
	{
		// stop condition
		if (isTerminal)
		{
			best_subtree = new MemeNode[1];
			best_subtree[0] = this;
			worst_subtree = new MemeNode[1];
			worst_subtree[0] = this;
			calculateLocalFitness(T);
			return;
		}
		
		// recursive
		leftChild.memetize(T);
		rightChild.memetize(T);
		
		// setting up the link lists:
		int size = (leftChild.best_subtree.length > rightChild.best_subtree.length?
				leftChild.best_subtree.length + 1:
				rightChild.best_subtree.length + 1);
		best_subtree = new MemeNode[size];
		worst_subtree = new MemeNode[size];
		
		best_subtree[0] = this;
		worst_subtree[0] = this;
		
		// assembling the best/worst arrays.
		int i = 1;
		while (i < size)
		{
			// if one of the subtrees is not long enough
			// copy the other one.
			if (rightChild.best_subtree.length < i)
			{
				best_subtree[i] = leftChild.best_subtree[i-1];
				worst_subtree[i] = leftChild.worst_subtree[i-1];
			}			
			else if (leftChild.best_subtree.length < i)
			{
				best_subtree[i] = rightChild.best_subtree[i-1];
				worst_subtree[i] = rightChild.worst_subtree[i-1];
			}
			// both subtrees are long enough - copy both.
			else
			{
				if (leftChild.best_subtree[i-1].fitness >
					rightChild.best_subtree[i-1].fitness)
					best_subtree[i] = leftChild.best_subtree[i-1];
				else
					best_subtree[i] = rightChild.best_subtree[i-1];
				
				if (leftChild.worst_subtree[i-1].fitness <
					rightChild.worst_subtree[i-1].fitness)
					worst_subtree[i] = leftChild.worst_subtree[i-1];
				else
					worst_subtree[i] = rightChild.worst_subtree[i-1];
			}
			
			i++;
		} // while for copying bestworst arrays
		
		/*
		 * Optimization function: simple hill climbing.
		 * 
		 * Each iteration, move the weight by meme_speed 
		 * to one direction. When the direction changes,
		 * meme_speed is divided by 5. when meme_speed 
		 * reaches meme_tresh, stops.
		 * 
		 * PS: Memetize effectively replaces the eval 
		 * function
		 * 
		 * TODO: Build a better optimization function.
		 */
		
		Parameter p = Parameter.getInstance();
		String param;
		
		param = p.getParam("meme speed");
		if (param == null)
		{
			System.err.println("Error: Meme Speed parameter not set");
			System.exit(1);
		}		
		double meme_speed = Double.parseDouble(param);
		
		param = p.getParam("meme tresh");
		if (param == null)
		{
			System.err.println("Error: Meme Tresh parameter not set");
			System.exit(1);
		}
		double meme_tresh = Double.parseDouble(param);
		
		param = p.getParam("meme delta");
		if (param == null)
		{
			System.err.println("Error: Meme Delta parameter not set");
			System.exit(1);
		}		
		double meme_delta = Double.parseDouble(param);
		int direction = 1;
		int count = 0;
		
		calculateLocalFitness(T);

		do
		{
			count++;
			
			double old_fit = fitness;
			weight = weight + (meme_speed*direction);
			calculateLocalFitness(T);

			// switch directions
			if (fitness < old_fit)
			{
				direction = direction * -1;
				meme_speed = meme_speed * meme_delta;
			}
			
			if (weight > (1.0 + meme_speed))
			{
				weight = 1.0;
				calculateLocalFitness(T);		
				break;
			}
			if (weight < (0.0 - meme_speed))
			{
				weight = 0.0;
				calculateLocalFitness(T);		
				break;
			}
			
		} while (meme_speed > meme_tresh);
		
	}
	
	

	/**
	 * This function should evaluate all non-sane nodes,
	 * generating their estimated returns, risk, and sharpe
	 * ratio. If the node is non-terminal, this function should
	 * recursively create a list of good nodes for 
	 * each level, based on the list of its two children.
	 */
	public void eval(Trainer T)
	{
		/* TODO: Sanity rules need to be reviewed
		 * Removing sanity makes thing slower, but less 
		 * prone to errors.
		 */
		//if (sane == true)
		//	return;
			
		
		// first recurse and create B/W subtree.
		if (isTerminal == false)
		{
			leftChild.eval(T);
			rightChild.eval(T);

			/* Creating Best/Worst Subtree. The
			 * zeroeth element is this tree.
			 */
			int size = (leftChild.best_subtree.length > rightChild.best_subtree.length?
					leftChild.best_subtree.length + 1:
					rightChild.best_subtree.length + 1);
			best_subtree = new MemeNode[size];
			worst_subtree = new MemeNode[size];
			
			best_subtree[0] = this;
			worst_subtree[0] = this;
			
			// assembling the best/worst arrays.
			int i = 1;
			while (i < size)
			{
				// if one of the subtrees is not long enough
				// copy the other one.
				if (rightChild.best_subtree.length < i)
				{
					best_subtree[i] = leftChild.best_subtree[i-1];
					worst_subtree[i] = leftChild.worst_subtree[i-1];
				}			
				else if (leftChild.best_subtree.length < i)
				{
					best_subtree[i] = rightChild.best_subtree[i-1];
					worst_subtree[i] = rightChild.worst_subtree[i-1];
				}
				// both subtrees are long enough - copy both.
				else
				{
					if (leftChild.best_subtree[i-1].fitness >
						rightChild.best_subtree[i-1].fitness)
						best_subtree[i] = leftChild.best_subtree[i-1];
					else
						best_subtree[i] = rightChild.best_subtree[i-1];
					
					if (leftChild.worst_subtree[i-1].fitness <
						rightChild.worst_subtree[i-1].fitness)
						worst_subtree[i] = leftChild.worst_subtree[i-1];
					else
						worst_subtree[i] = rightChild.worst_subtree[i-1];
				}
				
				i++;
			} // while for copying bestworst arrays
	
				
		}
		else //terminal case
		{
			best_subtree = new MemeNode[1];
			best_subtree[0] = this;
			worst_subtree = new MemeNode[1];
			worst_subtree[0] = this;
		}
		
		// now I finally evaluate the individual.
		calculateLocalFitness(T);
		
	}
	
	/**
	 * This function calculates the local fitness and 
	 * is called both by eval and by memetize
	 * @param T
	 */
	public void calculateLocalFitness(Trainer T)
	{
		// now I finally evaluate the individual.
		double[] port = generatePortfolio();
		Parameter p = Parameter.getInstance();
		String param;
		
		param = p.getParam("risk type");
		if (param == null)
		{
			System.err.println("Error. Parameter -risk type- not set");
			System.exit(0);
		}
		int running = param.compareToIgnoreCase("running");
		
		param = p.getParam("riskless return");
		if (param == null)
		{
			System.err.println("Error. Parameter \"riskless return\" not set");
			System.exit(0);
		}
		double r0 = Double.parseDouble(param);
		
		// calculate return and risk;
		Ereturn = T.getExpectedReturn(port);
		if (running == 0)
			Risk = T.getRunningRisk(port);
		else
			Risk = T.getCovarRisk(port);
		Sharpe = (Ereturn - r0)/Risk;
		
		/*
		 * TODO: Here I decide what the fitness 
		 * actually IS. I need to make this more complex
		 * to accomodate for other kinds of fitness.
		 */		
		fitness = Sharpe;
		
		sane = true;
		
	}
	
	/**
	 *  Generates a weight list for this tree 
	 */
	public double[] generatePortfolio()
	{
		Market mkt = Market.getInstance();

		double[] result = new double[mkt.assets.size()];
		
		if (isTerminal)
			{
				result[assetIndex] = 1.0;
			}
		else
		{
			double[] lresult = leftChild.generatePortfolio();
			double[] rresult = rightChild.generatePortfolio();	
			for (int i = 0; i < result.length; i++)
			{
					
				result[i] = (lresult[i] * weight) + (rresult[i] * (1 - weight));
			}
		}
		
		return result;
	}
		
	
	
	/**
	 *  implements comparable. Higher fitness is better.
     */
    public int compareTo(MemeNode gt)
    {
    	if (fitness > gt.fitness)
    		return -1;
    	if (fitness < gt.fitness)
    		return 1;
    	return 0;
    }
    
    /**
     * Counts the number of assets in this tree. Possibly
     * limited by a weight treshold.
     * 
     */
    public int countAsset(double tresh)
    {
    	int ret = 0;
    	double[] w = generatePortfolio();
    	for (int i = 0; i < w.length; i++)
    		if (w[i] > tresh) 
    			ret++;
    	return ret;    		
    }

    
    /**
     * Returns a text string that contains an ASCII 
     * representation of this tree. verbose, if set 
     * to true, prints each node's contents.
     */
    public String dumptree(boolean verbose)
    {
    	return dumptree(0,verbose);
    }
    public String dumptree(int tab, boolean verbose)
    {
    	String ret = "";
    	String tabulation = "";
    	for (int i = 0; i < tab; i++)
    		tabulation = tabulation + "  ";
    	
    	ret = ret + tabulation + "(Node " + uniqueID + ")";
    	if (isTerminal)
    	{
    		ret = ret + " Terminal: " + assetIndex;
    	}
    	else
    	{
    		ret = ret + " Non-Term: " + weight;
    	}

    	if (verbose)
    	{
    		ret = ret + "  Fitness: " + fitness;
    		if (mydepth == 0)
    		{
    			ret = ret + "\nBest: ";
    		for (int i = 0; i < this.best_subtree.length; i++)
    			ret = ret + best_subtree[i].uniqueID + " ";
    		
    		ret = ret + " Worst: ";
    		for (int i = 0; i < this.best_subtree.length; i++)
    			ret = ret + worst_subtree[i].uniqueID + " ";
    		}
    	}
    	
    	ret = ret + "\n";
    	
    	if (!isTerminal)
    	{
    		ret = ret + tabulation + leftChild.dumptree(tab + 1, verbose);
    		ret = ret + tabulation + rightChild.dumptree(tab + 1, verbose);
    	}
    		
    	return ret;
    }
	
    /**
     * Count the number of nodes and introns in this 
     * tree. Recursive. Introns are defined as the number 
     * of nodes that don't contribute to the final answer, 
     * because they are under zero weight.
     *  
     * @return
     * ret[0] is the total number of nodes, and 
     * ret[1] is the total number of introns;
     */
    public int[] countNodes()
    {
    	int ret[] = new int[2];
    	
    	if (isTerminal)
    	{
    		ret[0] = 1;
    		ret[1] = 0;
    	}
    	else
    	{
    		int lret[] = leftChild.countNodes();
    		int rret[] = rightChild.countNodes();
    		
    		ret[0] = lret[0] + rret[0] + 1;
    		
    		if (weight == 1.0)
    		{
    			ret[1] = lret[1] + rret[0];
    		}
    		else if (weight == 0.0)
    		{
    			ret[1] = lret[0] + rret[1];
    		}
    		else 
    		{
    			ret[1] = lret[1] + rret[1];
    		}
    		
    	}
    	
    	return ret;
    }
    
    public boolean insanity_Check()
    {
    	boolean ret = true;
    	
    	if (isTerminal)
    	{
    		return true;
    	}
    	
    	if (leftChild == null || rightChild == null)
    	{
    		return false;
    	}
    	
    	ret = leftChild.insanity_Check() && ret; 
    	ret = rightChild.insanity_Check() && ret;
    	
    	return ret;
    }
    
    
}
